/*
	@file
	simplesensel
	First Attempt to use Sensel Morph in Max 7/64 bits / Windows
	@ingroup	Roald Baudoux

	Global description: object to connect to a Sensel Morph device, control it and get its data
	Uses Sensel C API and Max SDK 7.3.3.

	2017-08-16	Beta 1 - All things related to multithreading are frozen for now. No more time to develop currently.
				Todo:
				- prevent close function to crash object
				- develop "open by com port"
				- allow to get firmware and sensor data from any device, not just the last one opened (by ID)
				- optimize (multithreading?, internal clock?, contacts if/switch simplification)
				- check why LEDs are not switched off when individuals contacts stopped
*/

#include "ext.h"							// standard Max include, always required
#include "ext_obex.h"						// required for new style Max object

#include <stdio.h>							// These lines required by Sensel
#include <stdlib.h>
#include <string.h>
#include "sensel.h"							// Sensel stuff
#include "sensel_device.h"
#include <math.h>



const unsigned int	max_contacts = 32;	// maximum number of contacts - should be extracted from each device in future versions

//============================== Max object struct
typedef struct _simplesensel
{
	t_object			ob;					// the object itself (must be first)
	void				*proxy;				// proxy inlet
	long				proxy_inletnum;		// # of inlet currently in use
	void				*outlet0;
	void				*outlet1;
	void				*outlet2;
	void				*outlet3;
	void				*outlet4;
	void				*outlet5;
	void				*outlet6;
	void				*outlet7;
	void				*outlet8;
	void				*outlet9;
	void				*outlet10;
	void				*outlet11;
	void				*outlet12;
	unsigned char		current_device;		// ID of device currently open
	signed short		device_is_open;		// switch of activation for contact data transmission
	//char				a_takecontacts;		// attribute 
	char				a_takeaccel;		// attribute
	char				a_takeforces;		// attribute
	char				a_takepositions;	// attribute
	//char				a_takelabels;		// attribute
	char				a_takeellipses;		// attribute
	char				a_takedeltas;		// attribute
	char				a_takeboundingboxes;// attribute
	char				a_takepeaks;		// attribute
	//int				a_sleeptime;		// attribute - how many milliseconds to sleep between data seeking operations
	unsigned char		a_mode;				// attribute - contacts output mode

	// multithreading stuff
	t_systhread			x_systhread;		// thread reference
	t_systhread_mutex	x_mutex;			// mutual exclusion lock for threadsafety
	int					x_systhread_cancel;	// thread cancel flag - not required anymore?
	void				*x_qelem;			// for message passing between threads
	int					active;				// data seeking switch

	// storage for data transmission with secondary thread
	t_atom				x_myContactData[20];	// for mode 0
	t_atom*				x_myContacts;		// for mode 1...
	t_atom*				x_myForces;
	t_atom*				x_myPositions;
	t_atom*				x_myEllipses;
	t_atom*				x_myBoundingboxes;
	t_atom*				x_myDeltasL;
	t_atom*				x_myPeaks;
	int					x_myContactDataSize;

	// =============== Get data from Sensel Morph(s)	
	SENSEL_HANDLE handle;		//Handle that references a Sensel device
	SenselDeviceList list;				//List of all available Sensel devices
	SenselFirmwareInfo fw_info;			//Firmware info from the Sensel device
	SenselSensorInfo sensor_info;		//Sensor info from the Sensel device
	SenselFrameData *frame;		//SenselFrame data that will hold the contacts

} t_simplesensel;

///////////////////////// function prototypes
// base set
void *rb_simplesensel_new(t_symbol *s, long argc, t_atom *argv);
void rb_simplesensel_free(t_simplesensel *x);
void rb_simplesensel_assist(t_simplesensel *x, void *b, long m, long a, char *s);
void rb_simplesensel_bang(t_simplesensel *x);
void rb_simplesensel_stop(t_simplesensel *x);
void rb_simplesensel_int(t_simplesensel *x, long n);
void rb_simplesensel_contacts(t_simplesensel *x);
void rb_simplesensel_getfirmwareversion(t_simplesensel *x);
void rb_simplesensel_getsensorinfo(t_simplesensel *x);
void rb_simplesensel_getnumberofdevices(t_simplesensel *x);
void rb_simplesensel_getlistofdevices(t_simplesensel *x);
void rb_simplesensel_openbyid(t_simplesensel *x, t_symbol *s, long argc, t_atom *argv);
//void rb_simplesensel_openbycomport(t_simplesensel *x, t_symbol *s, long argc, t_atom *argv);
void rb_simplesensel_openfirst(t_simplesensel *x);
void rb_simplesensel_close(t_simplesensel *x);
//void rb_simplesensel_getmaxframerate(t_simplesensel *x);
//void rb_simplesensel_setmaxframerate(t_simplesensel *x, t_symbol *s, long argc, t_atom *argv);
//void rb_simplesensel_getscandetail(t_simplesensel *x);
//void rb_simplesensel_setscandetail(t_simplesensel *x, t_symbol *s, long argc, t_atom *argv);
//void rb_simplesensel_test(t_simplesensel *x);

// multithreading
//void *rb_simplesensel_threadproc(t_simplesensel *x);
//void rb_simplesensel_qfn(t_simplesensel *x);
//void rb_simplesensel_sleeptime(t_simplesensel *x, long sleeptime);

//////////////////////// global class pointer variable
void *rb_simplesensel_class;

//=========================================== METHODS ====================================================================

// method run upon first object creation

void ext_main(void *r)
{
	t_class *c;

	c = class_new("rb.simplesensel", (method)rb_simplesensel_new, (method)rb_simplesensel_free, (long)sizeof(t_simplesensel),
		0L /* leave NULL!! */, A_GIMME, 0);

	/* Methods declaration */
	class_addmethod(c, (method)rb_simplesensel_assist, "assist", A_CANT, 0);
	class_addmethod(c, (method)rb_simplesensel_bang, "bang", 0);
	class_addmethod(c, (method)rb_simplesensel_int, "int", A_LONG, 0);
	class_addmethod(c, (method)rb_simplesensel_contacts, "contacts", A_CANT, 0);
	class_addmethod(c, (method)rb_simplesensel_getfirmwareversion, "getfirmwareversion", 0);
	class_addmethod(c, (method)rb_simplesensel_getsensorinfo, "getsensorinfo", 0);
	class_addmethod(c, (method)rb_simplesensel_getnumberofdevices, "getnumberofdevices", 0);           
	class_addmethod(c, (method)rb_simplesensel_getlistofdevices, "getlistofdevices", 0);
	class_addmethod(c, (method)rb_simplesensel_openbyid, "openbyid", A_GIMME, 0);
	//class_addmethod(c, (method)rb_simplesensel_openbycomport, "openbycomport", A_GIMME, 0);
	class_addmethod(c, (method)rb_simplesensel_openfirst, "openfirst", 0);
	class_addmethod(c, (method)rb_simplesensel_close, "close", 0);
	//class_addmethod(c, (method)rb_simplesensel_getmaxframerate, "getmaxframerate", 0);
	//class_addmethod(c, (method)rb_simplesensel_setmaxframerate, "setmaxframerate", A_GIMME, 0);
	//class_addmethod(c, (method)rb_simplesensel_sleeptime, "sleeptime", A_DEFLONG, 0);
	//class_addmethod(c, (method)rb_simplesensel_test, "test", 0);
	//class_addmethod(c, (method)rb_simplesensel_getscandetail, "getscandetail", 0);
	//class_addmethod(c, (method)rb_simplesensel_setscandetail, "setscandetail", A_GIMME, 0);

	// Attributes
	CLASS_ATTR_CHAR(c, "mode", 0, t_simplesensel, a_mode);
	//CLASS_ATTR_CHAR(c, "takecontacts", 0, t_simplesensel, a_takecontacts);
	CLASS_ATTR_CHAR(c, "takeaccel", 0, t_simplesensel, a_takeaccel);
	CLASS_ATTR_CHAR(c, "takeforces", 0, t_simplesensel, a_takeforces);
	CLASS_ATTR_CHAR(c, "takepositions", 0, t_simplesensel, a_takepositions);
	//CLASS_ATTR_CHAR(c, "takelabels", 0, t_simplesensel, a_takelabels);
	CLASS_ATTR_CHAR(c, "takeellipses", 0, t_simplesensel, a_takeellipses);
	CLASS_ATTR_CHAR(c, "takedeltas", 0, t_simplesensel, a_takedeltas);
	CLASS_ATTR_CHAR(c, "takeboundingboxes", 0, t_simplesensel, a_takeboundingboxes);
	CLASS_ATTR_CHAR(c, "takepeaks", 0, t_simplesensel, a_takepeaks);
	//CLASS_ATTR_CHAR(c, "sleeptime", 0, t_simplesensel, a_sleeptime);

	/* style */  /* label */
	CLASS_ATTR_STYLE_LABEL(c, "mode", 0, "enumindex", "Contacts data output mode");
	CLASS_ATTR_ENUMINDEX(c, "mode", 0, "\"Per contact\" \"Per data type\"");

	//CLASS_ATTR_STYLE_LABEL(c, "takecontacts", 0, "onoff", "Take contacts");
	CLASS_ATTR_STYLE_LABEL(c, "takeaccel", 0, "onoff", "Take accelerometer");
	CLASS_ATTR_STYLE_LABEL(c, "takeforces", 0, "onoff", "Take forces");
	CLASS_ATTR_STYLE_LABEL(c, "takepositions", 0, "onoff", "Take positions");
	//CLASS_ATTR_STYLE_LABEL(c, "takelabels", 0, "onoff", "Take labels");
	CLASS_ATTR_STYLE_LABEL(c, "takeellipses", 0, "onoff", "Take ellipses");
	CLASS_ATTR_STYLE_LABEL(c, "takedeltas", 0, "onoff", "Take deltas");
	CLASS_ATTR_STYLE_LABEL(c, "takeboundingboxes", 0, "onoff", "Take bounding boxes");
	CLASS_ATTR_STYLE_LABEL(c, "takepeaks", 0, "onoff", "Take peaks");
	//CLASS_ATTR_LABEL(c, "sleeptime", 0, "Sleep time");
	//CLASS_ATTR_FILTER_MIN(c, "sleeptime", 2); // minimum value set to 2 - no maximum

	class_register(CLASS_BOX, c); /* CLASS_NOBOX */
	rb_simplesensel_class = c;

	finder_addclass("Roald Baudoux", "rb.simplesensel");
	post("rb.simplesensel external - version beta 1 - 2017-08-16\n");
	post("compiled on %s", __DATE__);
	post("Roald Baudoux - http://www.roaldbaudoux.org\n");
	post("By the way: I want my Pluggo back!\n");
}

void *rb_simplesensel_new(t_symbol *s, long argc, t_atom *argv)
{
	t_simplesensel *x = NULL;

	if ((x = (t_simplesensel *)object_alloc(rb_simplesensel_class))) {
		// inlet creation
		//x->proxy = proxy_new(x, 1, &x->proxy_inletnum);	// fully-flexible inlet for any type

		// outlet creation
		x->outlet0 = outlet_new((t_object *)x, NULL); // generic outlet
		x->outlet1 = outlet_new((t_object *)x, NULL); // generic outlet
		x->outlet2 = outlet_new((t_object *)x, NULL); // generic outlet
		x->outlet3 = outlet_new((t_object *)x, NULL); // generic outlet
		x->outlet4 = outlet_new((t_object *)x, NULL); // generic outlet
		x->outlet5 = outlet_new((t_object *)x, NULL); // generic outlet
		x->outlet6 = outlet_new((t_object *)x, NULL); // generic outlet - accelerometer
		x->outlet7 = outlet_new((t_object *)x, NULL); // generic outlet	- ellipses
		x->outlet8 = outlet_new((t_object *)x, NULL); // generic outlet - deltas
		x->outlet9 = outlet_new((t_object *)x, NULL); // generic outlet - bounding boxes
		x->outlet10 = outlet_new((t_object *)x, NULL); // generic outlet - peaks
		x->outlet11 = outlet_new((t_object *)x, NULL); // generic outlet
		x->outlet12 = outlet_new((t_object *)x, NULL); // generic outlet
		x->current_device = 0xFF; // 0xFF means no connection established
		x->device_is_open = 0;

		// multithreading stuff
		//x->x_qelem = qelem_new(x, (method)rb_simplesensel_qfn);
		//x->x_systhread = NULL;
		//systhread_mutex_new(&x->x_mutex, 0);

		// attributes initialization
		x->a_mode = 0;				// "Output per contact" mode as default
		//x->a_takecontacts = 1;
		x->a_takepositions = 1;
		x->a_takeaccel = 1;
		x->a_takeforces = 1;
		//x->a_takelabels = 0;
		x->a_takeellipses = 0;
		x->a_takedeltas = 0;
		x->a_takeboundingboxes = 0;
		x->a_takepeaks = 0;

		//x->a_sleeptime = 50;

		// process attributes encoded in object box
		attr_args_process(x, argc, argv);

		x->handle = NULL;		//Handle that references a Sensel device
		x->frame = NULL;		//SenselFrame data that will hold the contacts

	}
	return (x);
}

void rb_simplesensel_assist(t_simplesensel *x, void *b, long m, long a, char *s){
	if (m == ASSIST_INLET) { // inlet
		switch (a) {
		case 0:
			sprintf(s, "send all messages here.");
			break;
		}
	}
	else {	// outlet
		switch (a) {
		case 0:
			sprintf(s, "outputs data in mode 0 - one list per contact with all requested data.");
			break;
		case 1:
			sprintf(s, "outputs accelerometer data as a triplet of x, y and z data (floats).");
			break;
		case 2:
			sprintf(s, "outputs peaks (floats)");
			break;
		case 3:
			sprintf(s, "outputs list of bounding boxes (min x, min y, max x, max y) (floats)");
			break;
		case 4:
			sprintf(s, "outputs list of deltas (floats)");
			break;
		case 5:
			sprintf(s, "outputs list of ellipses (float).");
			break;
		case 6:
			sprintf(s, "outputs list of forces as couples of touch ID (int) and force (float).");
			break;
		case 7:
			sprintf(s, "outputs global force (float).");
			break;
		case 8:
			sprintf(s, "outputs list of positions of contacts as triplets of touch ID (int), x position (float) and y position (float).");
			break;
		case 9:
			sprintf(s, "outputs list of contacts as couples of touch ID and state (ints).");
			break;
		case 10:
			sprintf(s, "outputs number of contacts (int).");
			break;
		case 11:
			sprintf(s, "outputs current state of connected device.");
			break;
		case 12:
			sprintf(s, "outputs various messages as triggered by 'get...' messages in inlets.");
			break;
		}
	}
}

void rb_simplesensel_free(t_simplesensel *x)
{
	object_free(x->proxy);
}

void rb_simplesensel_bang(t_simplesensel *x)
{
	if (x->current_device != 0xFF) {
		if (x->device_is_open == 1) {
			rb_simplesensel_contacts(x);
		}
		else error("No Sensel Morph open. Can't get data.\n");
	}
	else error("Can't get data. No Sensel Morph connected to this object.\n");	
}

// use an int to start and end data transmission
void rb_simplesensel_int(t_simplesensel *x, long n)
{
	post("This function is not really useful currently.");
/*	
	bool test = (n != 0);
	switch (test)
	{
	case false:
		if (x->active == 0) error("No contact data transmission to stop...\n");
		else
		{
			x->active = 0;
			post("Transmission of contact data stopped.\n");

			rb_simplesensel_stop(x);
			x->active = 0;
			post("Thread stopped.\n");
		}
		break;
	case true:
		if (x->active == 1) error("Contact data transmission already active...\n");
		else {
			//systhread_create((method)rb_simplesensel_threadproc, x, 0, 0, 0, &x->x_systhread);
			x->active = 1;
			post("Transmission of contact data activated.");
		}
		break;
	}
	*/
}

void rb_simplesensel_stop(t_simplesensel *x)
{
	unsigned int ret;

	if (x->x_systhread) {
		post("stopping our thread");
		x->x_systhread_cancel = true;						// tell the thread to stop
		systhread_join(x->x_systhread, &ret);					// wait for the thread to stop
		x->x_systhread = NULL;
	}
}

void rb_simplesensel_contacts(t_simplesensel *x)
{
	char forces_sem = 0, positions_sem = 0, accel_sem = 0, labels_sem = 0, ellipses_sem = 0, deltas_sem = 0, peaks_sem = 0, boundingboxes_sem = 0;
	forces_sem = x->a_takeforces;
	positions_sem = x->a_takepositions;
	accel_sem = x->a_takeaccel;
	//labels_sem = x->a_takelabels;
	ellipses_sem = x->a_takeellipses;
	deltas_sem = x->a_takedeltas;
	peaks_sem = x->a_takepeaks;
	boundingboxes_sem = x->a_takeboundingboxes;
	// Set x->frame mask and contact masks according to attributes
	char myFrameMask = 0x00;
	char myContactMask = 0x00;
	unsigned int num_frames = 0;

	switch (x->a_mode) {
	// Per contact mode : all data are sent together on leftmost output
	case 0:
		myFrameMask = myFrameMask + FRAME_CONTENT_CONTACTS_MASK;
		if (accel_sem == 1) myFrameMask = myFrameMask + FRAME_CONTENT_ACCEL_MASK;
		//if () myFrameMask = myFrameMask + FRAME_CONTENT_PRESSURE_MASK;
		//if () myFrameMask = myFrameMask + FRAME_CONTENT_LABELS_MASK;
		if (ellipses_sem == 1) myContactMask = myContactMask + CONTACT_MASK_ELLIPSE;
		if (deltas_sem == 1) myContactMask = myContactMask + CONTACT_MASK_DELTAS;
		if (boundingboxes_sem == 1) myContactMask = myContactMask + CONTACT_MASK_BOUNDING_BOX;
		if (peaks_sem == 1) myContactMask = myContactMask + CONTACT_MASK_PEAK;

		senselSetFrameContent(x->handle, myFrameMask);
		//Allocate a x->frame of data, must be done before reading x->frame data
		senselAllocateFrameData(x->handle, &x->frame);
		//Set data to include in contacts
		senselSetContactsMask(x->handle, myContactMask);
		//Start scanning the Sensel device
		senselStartScanning(x->handle);

		unsigned int num_frames = 0;
		//Read all available data from the Sensel device
		senselReadSensor(x->handle);
		/*
		switch (senselReadSensor(x->handle)) {
		case (0):
		post("Read sensor: success.\n");
		break;
		case (-1):
		post("Read sensor : failure.\n");
		};
		*/
		//Get number of x->frames available in the data read from the sensor
		senselGetNumAvailableFrames(x->handle, &num_frames);
		//post("Number of available x->frames: %d\n", num_frames);
		for (unsigned int f = 0; f < num_frames; f++)
		{
			//Read one x->frame of data
			senselGetFrame(x->handle, x->frame);
			//Print out contact data
			outlet_int(x->outlet2, x->frame->n_contacts);
			// stores global force

			if(x->frame->n_contacts > 0) {
				int numberOfData = 2; // number of items to transmit - 2 t_atom minimum (id and state)
				t_atom* myContactData = NULL;
				if (positions_sem == 1) numberOfData = numberOfData + 2;
				if (forces_sem == 1) numberOfData = numberOfData + 1;
				if (ellipses_sem == 1) numberOfData = numberOfData + 3;
				if (deltas_sem == 1) numberOfData = numberOfData + 4;
				if (boundingboxes_sem == 1) numberOfData = numberOfData + 4;
				if (peaks_sem == 1) numberOfData = numberOfData + 4;
				myContactData = malloc(numberOfData * sizeof(t_atom));
				if (myContactData != NULL) {
					double globalForce = 0.0;
					for (int c = 0; c < x->frame->n_contacts; c++)
					{
						unsigned int state = x->frame->contacts[c].state;
						unsigned int arrayposition = 1;
						// fill array with contact ID and state
						atom_setlong(myContactData, x->frame->contacts[c].id);
						atom_setlong(myContactData + arrayposition, x->frame->contacts[c].state);
						arrayposition += 1;
						// fill array with positions
						if (positions_sem == 1) {
							atom_setfloat(myContactData + arrayposition, x->frame->contacts[c].x_pos);
							atom_setfloat(myContactData + arrayposition + 1, x->frame->contacts[c].y_pos);
							arrayposition += 2;
						}
						// fill array with forces				
						if (forces_sem == 1) {
							atom_setfloat(myContactData + arrayposition, x->frame->contacts[c].total_force);
							arrayposition += 1;
							globalForce = globalForce + x->frame->contacts[c].total_force;
						}
						// fill array with ellipses
						if (ellipses_sem == 1) {
							atom_setfloat(myContactData + arrayposition, x->frame->contacts[c].orientation);
							atom_setfloat(myContactData + arrayposition + 1, x->frame->contacts[c].major_axis);
							atom_setfloat(myContactData + arrayposition + 2, x->frame->contacts[c].minor_axis);
							arrayposition += 3;
						}
						// fill array with deltas
						if (deltas_sem == 1) {
							atom_setfloat(myContactData + arrayposition, x->frame->contacts[c].delta_x);
							atom_setfloat(myContactData + arrayposition + 1, x->frame->contacts[c].delta_y);
							atom_setfloat(myContactData + arrayposition + 2, x->frame->contacts[c].delta_force);
							atom_setfloat(myContactData + arrayposition + 3, x->frame->contacts[c].delta_area);
							arrayposition += 4;
						}
						// fill array with bounding boxes
						if (boundingboxes_sem == 1) {
							atom_setfloat(myContactData + arrayposition, x->frame->contacts[c].min_x);
							atom_setfloat(myContactData + arrayposition + 1, x->frame->contacts[c].min_y);
							atom_setfloat(myContactData + arrayposition + 2, x->frame->contacts[c].max_x);
							atom_setfloat(myContactData + arrayposition + 3, x->frame->contacts[c].max_y);
							arrayposition += 4;
						}
						// fill array with peaks
						if (peaks_sem == 1) {
							atom_setfloat(myContactData + arrayposition, x->frame->contacts[c].peak_x);
							atom_setfloat(myContactData + arrayposition + 1, x->frame->contacts[c].peak_y);
							atom_setfloat(myContactData + arrayposition + 2, x->frame->contacts[c].peak_force);
							// arrayposition += 3; // to be uncommented only if additional data are addded
						}

						//Turn on LED for CONTACT_START
						if (state == CONTACT_START) {
							senselSetLEDBrightness(x->handle, x->frame->contacts[c].id, 100);
						}
						else if (state == CONTACT_MOVE) {
							//
						}
						else senselSetLEDBrightness(x->handle, x->frame->contacts[c].id, 0);
						/*
						//Turn off LED for CONTACT_END
						else if (state == CONTACT_END) {
							senselSetLEDBrightness(x->handle, x->frame->contacts[c].id, 0);
						}
						*/

						// output data
						outlet_list(x->outlet12, NULL, numberOfData, myContactData);
					}
					free(myContactData);
					if (forces_sem == 1) outlet_float(x->outlet5, globalForce);

					//Read accel data
					if (x->a_takeaccel == 1) {
						t_atom myAccelData[3];
						atom_setlong(myAccelData, x->frame->accel_data->x);
						atom_setlong(myAccelData + 1, x->frame->accel_data->y);
						atom_setlong(myAccelData + 2, x->frame->accel_data->z);
						outlet_list(x->outlet11, NULL, 3, myAccelData);
					}
					//else exit(0);
				}
			}
		}
		break;
	// Per type mode : each data type is sent on an independant outlet
	case 1:
		myFrameMask = myFrameMask + FRAME_CONTENT_CONTACTS_MASK;
		if (accel_sem == 1) myFrameMask = myFrameMask + FRAME_CONTENT_ACCEL_MASK;
		//if () myFrameMask = myFrameMask + FRAME_CONTENT_PRESSURE_MASK;
		//if () myFrameMask = myFrameMask + FRAME_CONTENT_LABELS_MASK;
		if (ellipses_sem == 1) myContactMask = myContactMask + CONTACT_MASK_ELLIPSE;
		if (deltas_sem == 1) myContactMask = myContactMask + CONTACT_MASK_DELTAS;
		if (boundingboxes_sem == 1) myContactMask = myContactMask + CONTACT_MASK_BOUNDING_BOX;
		if (peaks_sem == 1) myContactMask = myContactMask + CONTACT_MASK_PEAK;

		senselSetFrameContent(x->handle, myFrameMask);
		//Allocate a x->frame of data, must be done before reading x->frame data
		senselAllocateFrameData(x->handle, &x->frame);
		//Set data to include in contacts
		senselSetContactsMask(x->handle, myContactMask);
		//Start scanning the Sensel device
		senselStartScanning(x->handle);
		//Read all available data from the Sensel device
		senselReadSensor(x->handle);
		/*
		switch (senselReadSensor(x->handle)) {
		case (0):
		post("Read sensor: success.\n");
		break;
		case (-1):
		post("Read sensor : failure.\n");
		};
		*/
		//Get number of x->frames available in the data read from the sensor
		senselGetNumAvailableFrames(x->handle, &num_frames);
		//post("Number of available x->frames: %d\n", num_frames);
		for (unsigned int f = 0; f < num_frames; f++)
		{
			//Read one x->frame of data
			senselGetFrame(x->handle, x->frame);
			//Print out contact data
			outlet_int(x->outlet2, x->frame->n_contacts);
			// stores global force

			if(x->frame->n_contacts > 0) {
				//post("Num Contacts: %d\n", x->frame->n_contacts);
				t_atom* myContacts = NULL;
				t_atom* myForces = NULL;
				t_atom* myPositions = NULL;
				t_atom* myEllipses = NULL;
				t_atom* myBoundingboxes = NULL;
				t_atom* myDeltas = NULL;
				t_atom* myPeaks = NULL;
				myContacts = malloc(( x->frame->n_contacts * 2) * sizeof(t_atom));
				if (positions_sem == 1) {
					myPositions = malloc(( x->frame->n_contacts * 3) * sizeof(t_atom));
				}
				if (forces_sem == 1) {
					myForces = malloc(( x->frame->n_contacts * 2) * sizeof(t_atom));
				}
				if (ellipses_sem == 1) {
					myEllipses = malloc(( x->frame->n_contacts * 4) * sizeof(t_atom));
				}
				if (deltas_sem == 1) {
					myDeltas = malloc(( x->frame->n_contacts * 5) * sizeof(t_atom));
				}
				if (boundingboxes_sem == 1) {
					myBoundingboxes = malloc(( x->frame->n_contacts * 5) * sizeof(t_atom));
				}
				if (peaks_sem == 1) {
					myPeaks = malloc(( x->frame->n_contacts * 4) * sizeof(t_atom));
				}
				if (myContacts != NULL && myPositions != NULL) {
					double globalForce = 0.0;
					for (int c = 0; c < x->frame->n_contacts; c++)
					{
						unsigned int state = x->frame->contacts[c].state;
						// fill array with contacts
						atom_setlong(myContacts + (c * 2), x->frame->contacts[c].id);
						atom_setlong(myContacts + (c * 2) + 1, x->frame->contacts[c].state);
						// fill array with positions
						if (positions_sem == 1) {
							atom_setlong(myPositions + (c * 3), x->frame->contacts[c].id);
							atom_setfloat(myPositions + (c * 3) + 1, x->frame->contacts[c].x_pos);
							atom_setfloat(myPositions + (c * 3) + 2, x->frame->contacts[c].y_pos);
						}
						// fill array with forces				
						if (forces_sem == 1) {
							atom_setlong(myForces + (c * 2), x->frame->contacts[c].id);
							atom_setfloat(myForces + (c * 2) + 1, x->frame->contacts[c].total_force);
							globalForce = globalForce + x->frame->contacts[c].total_force;
						}
						// fill array with ellipses
						if (ellipses_sem == 1) {
							atom_setlong(myEllipses + (c * 4), x->frame->contacts[c].id);
							atom_setfloat(myEllipses + (c * 4) + 1, x->frame->contacts[c].orientation);
							atom_setfloat(myEllipses + (c * 4) + 2, x->frame->contacts[c].major_axis);
							atom_setfloat(myEllipses + (c * 4) + 3, x->frame->contacts[c].minor_axis);
						}
						// fill array with deltas
						if (deltas_sem == 1) {
							atom_setlong(myDeltas + (c * 5), x->frame->contacts[c].id);
							atom_setfloat(myDeltas + (c * 5) + 1, x->frame->contacts[c].delta_x);
							atom_setfloat(myDeltas + (c * 5) + 2, x->frame->contacts[c].delta_y);
							atom_setfloat(myDeltas + (c * 5) + 3, x->frame->contacts[c].delta_force);
							atom_setfloat(myDeltas + (c * 5) + 4, x->frame->contacts[c].delta_area);
						}
						// fill array with bounding boxes
						if (boundingboxes_sem == 1) {
							atom_setlong(myBoundingboxes + (c * 5), x->frame->contacts[c].id);
							atom_setfloat(myBoundingboxes + (c * 5) + 1, x->frame->contacts[c].min_x);
							atom_setfloat(myBoundingboxes + (c * 5) + 2, x->frame->contacts[c].min_y);
							atom_setfloat(myBoundingboxes + (c * 5) + 3, x->frame->contacts[c].max_x);
							atom_setfloat(myBoundingboxes + (c * 5) + 4, x->frame->contacts[c].max_y);
						}
						// fill array with peaks
						if (peaks_sem == 1) {
							atom_setlong(myPeaks + (c * 4), x->frame->contacts[c].id);
							atom_setfloat(myPeaks + (c * 4) + 1, x->frame->contacts[c].peak_x);
							atom_setfloat(myPeaks + (c * 4) + 2, x->frame->contacts[c].peak_y);
							atom_setfloat(myPeaks + (c * 4) + 3, x->frame->contacts[c].peak_force);
						}

						//Turn on LED for CONTACT_START
						if (state == CONTACT_START) {
							senselSetLEDBrightness(x->handle, x->frame->contacts[c].id, 100);
						}
						else if (state == CONTACT_MOVE) {
							//
						}
						else senselSetLEDBrightness(x->handle, x->frame->contacts[c].id, 0);
						/*
						//Turn off LED for CONTACT_END
						else if (state == CONTACT_END) {
						senselSetLEDBrightness(x->handle, x->frame->contacts[c].id, 0);
						}
						*/
					}
					// output data
					outlet_list(x->outlet3, NULL,(x->frame->n_contacts * 2), myContacts);
					if (positions_sem == 1) {
						outlet_list(x->outlet4, NULL,(x->frame->n_contacts * 3), myPositions);
						free(myPositions);
					}
					if (forces_sem == 1) {
						outlet_float(x->outlet5, globalForce);
						outlet_list(x->outlet6, NULL,(x->frame->n_contacts * 2), myForces);
						free(myForces);
					}
					if (ellipses_sem == 1) {
						outlet_list(x->outlet7, NULL,(x->frame->n_contacts * 4), myEllipses);
						free(myEllipses);
					}
					if (deltas_sem == 1) {
						outlet_list(x->outlet8, NULL,(x->frame->n_contacts * 5), myDeltas);
						free(myDeltas);
					}
					if (boundingboxes_sem == 1) {
						outlet_list(x->outlet9, NULL,(x->frame->n_contacts * 5), myBoundingboxes);
						free(myBoundingboxes);
					}
					if (peaks_sem == 1) {
						outlet_list(x->outlet10, NULL,(x->frame->n_contacts * 4), myPeaks);
						free(myPeaks);
					}
					free(myContacts);

					//Read accel data
					if (x->a_takeaccel == 1) {
						t_atom myAccelData[3];
						atom_setlong(myAccelData, x->frame->accel_data->x);
						atom_setlong(myAccelData + 1, x->frame->accel_data->y);
						atom_setlong(myAccelData + 2, x->frame->accel_data->z);
						outlet_list(x->outlet11, NULL, 3, myAccelData);
					}
					//else exit(0);
				}
			}
		}
		break;
	}
	senselStopScanning(x->handle);
}

// triggered by the helper thread
/*
void rb_simplesensel_qfn(t_simplesensel *x)
{
	//int myfoo;

	systhread_mutex_lock(x->x_mutex);
	//myfoo = x->x_foo;															// access shared data
	systhread_mutex_unlock(x->x_mutex);

	// *never* wrap outlet calls with systhread_mutex_lock()
	outlet_list(x->outlet12, NULL, x->x_myContactDataSize, x->x_myContactData);
	
}
*/

/*
void *rb_simplesensel_threadproc(t_simplesensel *x)
{
	char forces_sem = 0, positions_sem = 0, accel_sem = 0, labels_sem = 0, ellipses_sem = 0, deltas_sem = 0,
		peaks_sem = 0, boundingboxes_sem = 0, myFrameMask = 0x00, myContactMask = 0x00, currentMode = -1;
	int mySleeptime = 0;
	while (x->active == 1) {
		systhread_mutex_lock(x->x_mutex);
		forces_sem = x->a_takeforces;
		positions_sem = x->a_takepositions;
		accel_sem = x->a_takeaccel;
		labels_sem = x->a_takelabels;
		ellipses_sem = x->a_takeellipses;
		deltas_sem = x->a_takedeltas;
		peaks_sem = x->a_takepeaks;
		boundingboxes_sem = x->a_takeboundingboxes;
		mySleeptime = x->a_sleeptime;
		currentMode = x->a_mode;
		systhread_mutex_unlock(x->x_mutex);

		switch (currentMode) {
			// Per contact mode : all data are sent together on leftmost output
		case 0:
			myFrameMask = myFrameMask + x->frame_CONTENT_CONTACTS_MASK;
			if (accel_sem == 1) myFrameMask = myFrameMask + x->frame_CONTENT_ACCEL_MASK;
			//if () myFrameMask = myFrameMask + x->frame_CONTENT_PRESSURE_MASK;
			//if () myFrameMask = myFrameMask + x->frame_CONTENT_LABELS_MASK;
			if (ellipses_sem == 1) myContactMask = myContactMask + CONTACT_MASK_ELLIPSE;
			if (deltas_sem == 1) myContactMask = myContactMask + CONTACT_MASK_DELTAS;
			if (boundingboxes_sem == 1) myContactMask = myContactMask + CONTACT_MASK_BOUNDING_BOX;
			if (peaks_sem == 1) myContactMask = myContactMask + CONTACT_MASK_PEAK;

			senselSetFrameContent(x->handle, myFrameMask);
			//Allocate a x->frame of data, must be done before reading x->frame data
			senselAllocateFrameData(x->handle, &x->frame);
			//Set data to include in contacts
			senselSetContactsMask(x->handle, myContactMask);
			//Start scanning the Sensel device
			senselStartScanning(x->handle);

			unsigned int num_frames = 0;
			//Read all available data from the Sensel device
			senselReadSensor(x->handle);
			switch (senselReadSensor(x->handle)) {
			case (0):
			post("Read sensor: success.\n");
			break;
			case (-1):
			post("Read sensor : failure.\n");
			};
			//Get number of x->frames available in the data read from the sensor
			senselGetNumAvailableFrames(x->handle, &num_frames);
			//post("Number of available frames: %d\n", num_frames);
			for (unsigned int f = 0; f < num_frames; f++)
			{
				//Read one frame of data
				senselGetFrame(x->handle, x->frame);
				//Print out contact data
				outlet_int(x->outlet2, x->frame->n_contacts);
				// stores global force

				if(x->frame->n_contacts > 0) {
					int numberOfData = 2; // number of items to transmit - 2 t_atom minimum (id and state)
					t_atom* myContactData = NULL;
					if (positions_sem == 1) numberOfData = numberOfData + 2;
					if (forces_sem == 1) numberOfData = numberOfData + 1;
					if (ellipses_sem == 1) numberOfData = numberOfData + 3;
					if (deltas_sem == 1) numberOfData = numberOfData + 4;
					if (boundingboxes_sem == 1) numberOfData = numberOfData + 4;
					if (peaks_sem == 1) numberOfData = numberOfData + 4;
					myContactData = malloc(numberOfData * sizeof(t_atom));
					if (myContactData != NULL) {
						double globalForce = 0.0;
						for (int c = 0; c < x->frame->n_contacts; c++)
						{
							unsigned int state = x->frame->contacts[c].state;
							unsigned int arrayposition = 1;
							// fill array with contact ID and state
							atom_setlong(myContactData, x->frame->contacts[c].id);
							atom_setlong(myContactData + arrayposition, x->frame->contacts[c].state);
							arrayposition += 1;
							// fill array with positions
							if (positions_sem == 1) {
								atom_setfloat(myContactData + arrayposition, x->frame->contacts[c].x_pos);
								atom_setfloat(myContactData + arrayposition + 1, x->frame->contacts[c].y_pos);
								arrayposition += 2;
							}
							// fill array with forces				
							if (forces_sem == 1) {
								atom_setfloat(myContactData + arrayposition, x->frame->contacts[c].total_force);
								arrayposition += 1;
								globalForce = globalForce + x->frame->contacts[c].total_force;
							}
							// fill array with ellipses
							if (ellipses_sem == 1) {
								atom_setfloat(myContactData + arrayposition, x->frame->contacts[c].orientation);
								atom_setfloat(myContactData + arrayposition + 1, x->frame->contacts[c].major_axis);
								atom_setfloat(myContactData + arrayposition + 2, x->frame->contacts[c].minor_axis);
								arrayposition += 3;
							}
							// fill array with deltas
							if (deltas_sem == 1) {
								atom_setfloat(myContactData + arrayposition, x->frame->contacts[c].delta_x);
								atom_setfloat(myContactData + arrayposition + 1, x->frame->contacts[c].delta_y);
								atom_setfloat(myContactData + arrayposition + 2, x->frame->contacts[c].delta_force);
								atom_setfloat(myContactData + arrayposition + 3, x->frame->contacts[c].delta_area);
								arrayposition += 4;
							}
							// fill array with bounding boxes
							if (boundingboxes_sem == 1) {
								atom_setfloat(myContactData + arrayposition, x->frame->contacts[c].min_x);
								atom_setfloat(myContactData + arrayposition + 1, x->frame->contacts[c].min_y);
								atom_setfloat(myContactData + arrayposition + 2, x->frame->contacts[c].max_x);
								atom_setfloat(myContactData + arrayposition + 3, x->frame->contacts[c].max_y);
								arrayposition += 4;
							}
							// fill array with peaks
							if (peaks_sem == 1) {
								atom_setfloat(myContactData + arrayposition, x->frame->contacts[c].peak_x);
								atom_setfloat(myContactData + arrayposition + 1, x->frame->contacts[c].peak_y);
								atom_setfloat(myContactData + arrayposition + 2, x->frame->contacts[c].peak_force);
								// arrayposition += 3; // to be uncommented only if additional data are addded
							}

							//Turn on LED for CONTACT_START
							if (state == CONTACT_START) {
								senselSetLEDBrightness(x->handle, x->frame->contacts[c].id, 100);
							}
							//Turn off LED for CONTACT_END
							else if (state == CONTACT_END) {
								senselSetLEDBrightness(x->handle, x->frame->contacts[c].id, 0);
							}

							// transmit data to struct

							systhread_mutex_lock(x->x_mutex);
							x->x_myContactDataSize = numberOfData;
							for (int k = 0; k < numberOfData; k++) {
								x->x_myContactData[k] = myContactData[k];
							}
							systhread_mutex_unlock(x->x_mutex);

							//outlet_list(x->outlet12, NULL, numberOfData, myContactData);
						}
						free(myContactData);
						if (forces_sem == 1) outlet_float(x->outlet5, globalForce);

						//Read accel data
						if (x->a_takeaccel == 1) {
							t_atom myAccelData[3];
							atom_setlong(myAccelData, x->frame->accel_data->x);
							atom_setlong(myAccelData + 1, x->frame->accel_data->y);
							atom_setlong(myAccelData + 2, x->frame->accel_data->z);
							//outlet_list(x->outlet11, NULL, 3, myAccelData);
						}
					}
				}
			}
			break;
			// Per data type mode : each data type is sent on an independant outle
		case 1:
			myFrameMask = myFrameMask + x->frame_CONTENT_CONTACTS_MASK;
			if (accel_sem == 1) myFrameMask = myFrameMask + x->frame_CONTENT_ACCEL_MASK;
			//if () myFrameMask = myFrameMask + x->frame_CONTENT_PRESSURE_MASK;
			//if () myFrameMask = myFrameMask + x->frame_CONTENT_LABELS_MASK;
			if (ellipses_sem == 1) myContactMask = myContactMask + CONTACT_MASK_ELLIPSE;
			if (deltas_sem == 1) myContactMask = myContactMask + CONTACT_MASK_DELTAS;
			if (boundingboxes_sem == 1) myContactMask = myContactMask + CONTACT_MASK_BOUNDING_BOX;
			if (peaks_sem == 1) myContactMask = myContactMask + CONTACT_MASK_PEAK;

			senselSetFrameContent(x->handle, myFrameMask);
			//Allocate a x->frame of data, must be done before reading x->frame data
			senselAllocateFrameData(x->handle, &x->frame);
			//Set data to include in contacts
			senselSetContactsMask(x->handle, myContactMask);
			//Start scanning the Sensel device
			senselStartScanning(x->handle);
			//Read all available data from the Sensel device
			senselReadSensor(x->handle);
			/*
			switch (senselReadSensor(x->handle)) {
			case (0):
			post("Read sensor: success.\n");
			break;
			case (-1):
			post("Read sensor : failure.\n");
			};
			//Get number of x->frames available in the data read from the sensor
			senselGetNumAvailableFrames(x->handle, &num_frames);
			//post("Number of available x->frames: %d\n", num_frames);
			for (unsigned int f = 0; f < num_frames; f++)
			{
				//Read one x->frame of data
				senselGetFrame(x->handle, x->frame);
				//Print out contact data
				outlet_int(x->outlet2, x->frame->n_contacts);
				// stores global force

				if(x->frame->n_contacts > 0) {
					//post("Num Contacts: %d\n", x->frame->n_contacts);
					t_atom* myContacts = NULL;
					t_atom* myForces = NULL;
					t_atom* myPositions = NULL;
					t_atom* myEllipses = NULL;
					t_atom* myBoundingboxes = NULL;
					t_atom* myDeltas = NULL;
					t_atom* myPeaks = NULL;
					myContacts = malloc(( x->frame->n_contacts * 2) * sizeof(t_atom));
					if (positions_sem == 1) {
						myPositions = malloc(( x->frame->n_contacts * 3) * sizeof(t_atom));
					}
					if (forces_sem == 1) {
						myForces = malloc(( x->frame->n_contacts * 2) * sizeof(t_atom));
					}
					if (ellipses_sem == 1) {
						myEllipses = malloc(( x->frame->n_contacts * 4) * sizeof(t_atom));
					}
					if (deltas_sem == 1) {
						myDeltas = malloc(( x->frame->n_contacts * 5) * sizeof(t_atom));
					}
					if (boundingboxes_sem == 1) {
						myBoundingboxes = malloc(( x->frame->n_contacts * 5) * sizeof(t_atom));
					}
					if (peaks_sem == 1) {
						myPeaks = malloc(( x->frame->n_contacts * 4) * sizeof(t_atom));
					}
					if (myContacts != NULL && myPositions != NULL) {
						double globalForce = 0.0;
						for (int c = 0; c < x->frame->n_contacts; c++)
						{
							unsigned int state = x->frame->contacts[c].state;
							// fill array with contacts
							atom_setlong(myContacts + (c * 2), x->frame->contacts[c].id);
							atom_setlong(myContacts + (c * 2) + 1, x->frame->contacts[c].state);
							// fill array with positions
							if (positions_sem == 1) {
								atom_setlong(myPositions + (c * 3), x->frame->contacts[c].id);
								atom_setfloat(myPositions + (c * 3) + 1, x->frame->contacts[c].x_pos);
								atom_setfloat(myPositions + (c * 3) + 2, x->frame->contacts[c].y_pos);
							}
							// fill array with forces				
							if (forces_sem == 1) {
								atom_setlong(myForces + (c * 2), x->frame->contacts[c].id);
								atom_setfloat(myForces + (c * 2) + 1, x->frame->contacts[c].total_force);
								globalForce = globalForce + x->frame->contacts[c].total_force;
							}
							// fill array with ellipses
							if (ellipses_sem == 1) {
								atom_setlong(myEllipses + (c * 4), x->frame->contacts[c].id);
								atom_setfloat(myEllipses + (c * 4) + 1, x->frame->contacts[c].orientation);
								atom_setfloat(myEllipses + (c * 4) + 2, x->frame->contacts[c].major_axis);
								atom_setfloat(myEllipses + (c * 4) + 3, x->frame->contacts[c].minor_axis);
							}
							// fill array with deltas
							if (deltas_sem == 1) {
								atom_setlong(myDeltas + (c * 5), x->frame->contacts[c].id);
								atom_setfloat(myDeltas + (c * 5) + 1, x->frame->contacts[c].delta_x);
								atom_setfloat(myDeltas + (c * 5) + 2, x->frame->contacts[c].delta_y);
								atom_setfloat(myDeltas + (c * 5) + 3, x->frame->contacts[c].delta_force);
								atom_setfloat(myDeltas + (c * 5) + 4, x->frame->contacts[c].delta_area);
							}
							// fill array with bounding boxes
							if (boundingboxes_sem == 1) {
								atom_setlong(myBoundingboxes + (c * 5), x->frame->contacts[c].id);
								atom_setfloat(myBoundingboxes + (c * 5) + 1, x->frame->contacts[c].min_x);
								atom_setfloat(myBoundingboxes + (c * 5) + 2, x->frame->contacts[c].min_y);
								atom_setfloat(myBoundingboxes + (c * 5) + 3, x->frame->contacts[c].max_x);
								atom_setfloat(myBoundingboxes + (c * 5) + 4, x->frame->contacts[c].max_y);
							}
							// fill array with peaks
							if (peaks_sem == 1) {
								atom_setlong(myPeaks + (c * 4), x->frame->contacts[c].id);
								atom_setfloat(myPeaks + (c * 4) + 1, x->frame->contacts[c].peak_x);
								atom_setfloat(myPeaks + (c * 4) + 2, x->frame->contacts[c].peak_y);
								atom_setfloat(myPeaks + (c * 4) + 3, x->frame->contacts[c].peak_force);
							}

							//Turn on LED for CONTACT_START
							if (state == CONTACT_START) {
								senselSetLEDBrightness(x->handle, x->frame->contacts[c].id, 100);
							}
							//Turn off LED for CONTACT_END
							else if (state == CONTACT_END) {
								senselSetLEDBrightness(x->handle, x->frame->contacts[c].id, 0);
							}
						}
						// output data
						outlet_list(x->outlet3, NULL,(x->frame->n_contacts * 2), myContacts);
						if (positions_sem == 1) {
							outlet_list(x->outlet4, NULL,(x->frame->n_contacts * 3), myPositions);
							free(myPositions);
						}
						if (forces_sem == 1) {
							outlet_float(x->outlet5, globalForce);
							outlet_list(x->outlet6, NULL,(x->frame->n_contacts * 2), myForces);
							free(myForces);
						}
						if (ellipses_sem == 1) {
							outlet_list(x->outlet7, NULL,(x->frame->n_contacts * 4), myEllipses);
							free(myEllipses);
						}
						if (deltas_sem == 1) {
							outlet_list(x->outlet8, NULL,(x->frame->n_contacts * 5), myDeltas);
							free(myDeltas);
						}
						if (boundingboxes_sem == 1) {
							outlet_list(x->outlet9, NULL,(x->frame->n_contacts * 5), myBoundingboxes);
							free(myBoundingboxes);
						}
						if (peaks_sem == 1) {
							outlet_list(x->outlet10, NULL,(x->frame->n_contacts * 4), myPeaks);
							free(myPeaks);
						}
						free(myContacts);

						//Read accel data
						if (x->a_takeaccel == 1) {
							t_atom myAccelData[3];
							atom_setlong(myAccelData, x->frame->accel_data->x);
							atom_setlong(myAccelData + 1, x->frame->accel_data->y);
							atom_setlong(myAccelData + 2, x->frame->accel_data->z);
							outlet_list(x->outlet11, NULL, 3, myAccelData);
						}
						//else exit(0);
					}
				}
			}
			break;
		}
	}
	qelem_set(x->x_qelem);
	systhread_sleep(mySleeptime);
	return(0);
}
*/

void rb_simplesensel_getnumberofdevices(t_simplesensel *x)
{
	//Get the number of available Sensel devices
	senselGetDeviceList(&x->list);
	if (x->list.num_devices != 0) {	
		t_atom myNumberOfDevices[2];
		atom_setsym(myNumberOfDevices, gensym("numberofdevices"));
		atom_setlong(myNumberOfDevices + 1, x->list.num_devices);
		outlet_list(x->outlet0, NULL, 2, myNumberOfDevices);
	}
	else error("No Sensel Morph detected. Check connection(s).");
}

void rb_simplesensel_getlistofdevices(t_simplesensel *x)
{
	//Get a list of available Sensel devices (max. 16) with serial number, ID and serial (com) port used
	senselGetDeviceList(&x->list);
	if (x->list.num_devices != 0) {
		t_atom myListOfDevices[49];
		atom_setsym(myListOfDevices, gensym("listofdevices"));
		// IMPROVE MEMORY ALLOCATION ACCORDING TO REAL NUMBER OF DEVICES!!!
		short j;
		for (j = 0; j < x->list.num_devices; j++) {
			atom_setsym(myListOfDevices + (j * 3) + 1, gensym(x->list.devices[j].serial_num));
			atom_setlong(myListOfDevices + (j * 3) + 2, (long)x->list.devices[j].idx);
			atom_setsym(myListOfDevices + (j * 3) + 3, gensym(x->list.devices[j].com_port));
		}
		outlet_list(x->outlet0, NULL , j * 3 + 1, myListOfDevices);
	}
	else error("No Sensel Morph detected. Check connection(s).");
}

void rb_simplesensel_getfirmwareversion(t_simplesensel *x)
{
	if (x->current_device != 0xFF) {
		//Get the sensor Firmware version
		senselGetFirmwareInfo(x->handle, &x->fw_info);
		t_atom myFWdata[8];
		atom_setsym(myFWdata, gensym("firmwareversion"));
		atom_setlong(myFWdata + 1, x->fw_info.fw_protocol_version);
		atom_setlong(myFWdata + 2, x->fw_info.fw_version_major);
		atom_setlong(myFWdata + 3, x->fw_info.fw_version_minor);
		atom_setlong(myFWdata + 4, x->fw_info.fw_version_build);
		atom_setlong(myFWdata + 5, x->fw_info.fw_version_release);
		atom_setlong(myFWdata + 6, x->fw_info.device_id);
		atom_setlong(myFWdata + 7, x->fw_info.device_revision);
		outlet_list(x->outlet0, NULL, 8, myFWdata);
	}
	else error("Can't get firmware version. No Sensel Morph currently associated with this object.\n");
}

void rb_simplesensel_getsensorinfo(t_simplesensel *x)
{
	if (x->current_device != 0xFF) {
		//Get the sensor Size and Resolution
		senselGetSensorInfo(x->handle, &x->sensor_info);
		t_atom mySensorData[5];
		atom_setsym(mySensorData, gensym("sensorinfo"));
		atom_setfloat(mySensorData + 1, x->sensor_info.width);
		//post("sensor info width %f\n", x->sensor_info.width);
		atom_setfloat(mySensorData + 2, x->sensor_info.height);
		//post("sensor info height %f\n", x->sensor_info.height);
		atom_setlong(mySensorData + 3, x->sensor_info.num_cols);
		//post("sensor info num cols %d\n", x->sensor_info.num_cols);
		atom_setlong(mySensorData + 4, x->sensor_info.num_rows);
		//post("sensor info num rows %d\n", x->sensor_info.num_rows);
		outlet_list(x->outlet0, NULL, 5, mySensorData);
	}
	else error("Can't get sensor info. No Sensel Morph currently associated with this object.\n");
}

// open a device through its ID
void rb_simplesensel_openbyid(t_simplesensel *x, t_symbol *s, long argc, t_atom *argv)
{
	// Connects only if there's no open device associated with the object yet
	if (x->device_is_open == 0) {
		senselGetDeviceList(&x->list);
		if (argc > 0) {
			if (argv[0].a_type == A_LONG) {
				char myDevice = (char)atom_getlong(argv);
				short match = 0;
				for (int j = 0; j < x->list.num_devices; j++) {
					if (x->list.devices[j].idx == myDevice) {
						match = 1;
						if (senselOpenDeviceByID(&x->handle, x->list.devices[j].idx) == 0) {
							x->current_device = x->list.devices[j].idx;
							x->device_is_open = 1;
							long output = 1;
							outlet_int(x->outlet1, output);
							post("Sensel Morph with ID %d successfully opened.\n", myDevice);
						}
						else error("Couldn't open Sensel Morph with ID %d.\n", myDevice);
						break;
					}
				}
				if (match == 0) error("Can't find a Sensel Morph with ID %d. Please check value.\n", myDevice);
			}
			else error("'openbyid': wrong argument type. Should be an int.");
		}
		else error("'openbyid': please provide an ID (int) as argument.");
	}
	else error("There's already an open Sensel Morph (ID: %d) associated with this object.\n", x->current_device);
}

// open the first device listed whatever its ID
// (actually the ID matches the position in the list but Sensel might modify this in future APIs).
void rb_simplesensel_openfirst(t_simplesensel *x)
{
	senselGetDeviceList(&x->list);
	if (x->list.num_devices != 0) {
		if (x->current_device == 0xFF) {
			if (senselOpen(&x->handle) == 0) {
				x->current_device = x->list.devices[0].idx;
				x->device_is_open = 1;
				long output = 1;
				outlet_int(x->outlet1, output);
				post("First listed Sensel Morph (ID: %d) successfully opened.\n", x->list.devices[0].idx);
			}
			else error("Couldn't open first listed Sensel Morph.\n");
		}
		else error("Object is already associated with Sensel Morph with ID %d.\n", x->current_device);
	}
	else error("No Sensel Morph detected. Check connection(s).");
}

/*
void rb_simplesensel_openbycomport(t_simplesensel *x, t_symbol *s, long argc, t_atom *argv)
{
	senselGetDeviceList(&x->list);
	post("Number of arguments: %d.\n", argc);
	if (argc > 0) {
		if (argv[0].a_type == A_LONG  && atom_getlong(argv) > 0) {
			post("Value of argument: %d.\n", atom_getlong(argv));
			char my_com_port[2];
			t_symbol *my_sym;
			my_sym = atom_getsym(argv);
			strncpy_zero(my_com_port, my_sym->s_name, sizeof(my_com_port));
			short j;
			// scans all devices
			unsigned int comparison_flag = 0;
			for (j = 0; j < list.num_devices; j++) {
				post("j = %d\n", j);
				post("Value of argument: %d.\n", atom_getlong(argv));
				post("x->list.devices[j].com_port[3]: %c\n", x->list.devices[j].com_port[3]);
				post("Size of argument: %d\n", sizeof(atom_getlong(argv)));
				post("Size of x->list.devices[j].com_port[3]: %d\n", sizeof(x->list.devices[j].com_port[3]));
				if ((int64)(atom_getlong(argv)) == (int64)(x->list.devices[j].com_port[3])) {
					comparison_flag = 1;
					//break;
				}
			}
			post("Comparison flag: %d\n", comparison_flag);
			if (comparison_flag == 1) {
				post("Equality confirmed.\n");
				post("results %d\n", senselOpenDeviceByComPort(&x->handle, x->list.devices[j].com_port));
				if (senselOpenDeviceByComPort(&x->handle, list.devices[j].com_port) == 0) {
					post("Device with COM port %s successfully opened.\n", atom_getsym(argv));
				}
				else post("Couldn't open device with COM port %s.\n", atom_getsym(argv));
			}
			//else 
		}
		else post("Wrong argument type or value.\n");
	}
	else error("Please provide a serial (COM) port name as argument.\n");
}
*/

void rb_simplesensel_close(t_simplesensel *x)
{
	if (x->current_device != 0xFF && x->device_is_open == 1) {
		if (x->handle) {
			//post("Handle found.");
			if (senselClose(x->handle) == 0) {					
				post("Sensel Morph with ID %d successfully closed.\n", x->current_device);
				x->current_device = 0xFF;
				x->device_is_open = 0;
				long output = 0;
				outlet_int(x->outlet1, output);
			}
			else error("Couldn't close Sensel Morph with ID %d.\n", x->current_device);
		}
		else error("No handle found! Can't close Sensel Morph.");
	}
	else error("No Sensel Morph associated with this object to be closed.");
}

/*
void rb_simplesensel_getmaxframerate(t_simplesensel *x) 
// check if device open first?
{
	short maxFrameRate = 0;
	if (senselGetMaxFrameRate(x->handle, &maxFrameRate) == SENSEL_OK) {
		post("Device maximum x->frame rate: %d\n", maxFrameRate);
	}
	else error("Couldn't get maximum frame rate.");
}
*/

/*
void rb_simplesensel_setmaxframerate(t_simplesensel *x, t_symbol *s, long argc, t_atom *argv)
// check if device open first?
{
	if (x->current_device != 0xFF)
	{
		if (argc > 0) {
			post("Number of arguments: %d.\n", argc);
			if (argv[0].a_type == A_LONG) {
				if (atom_getlong(argv) > 0 && atom_getlong(argv) < 65536) {
					post("Value of argument: %d.\n", atom_getlong(argv));
					if (senselSetMaxFrameRate(x->handle, atom_getlong(argv)) == 0) {
						post("Max scan rate successfully set.\n");
					}
					else error("Couldn't set max scan rate.");
				}
				else post("Set max x->frame rate: wrong argument value.\n");
			}
			else post("Set max x->frame rate: wrong argument type.\n");
		}
		else error("Please provide an argument to the \"setmaxframerate\" message.\n");
	}
	else error("No Sensel Morph associated with this object.");
}
*/

/*
void rb_simplesensel_getscandetail(t_simplesensel *x)
{
	if (x->current_device != 0xFF)
	{
		SenselScanDetail scanDetail = 0;
		if (senselGetScanDetail(x->handle, &scanDetail) == SENSEL_OK) {
			post("Device scan detail: %d\n", scanDetail);
		}
		else error("Couldn't get scan detail.");
	}
	else error("No Sensel Morph associated with this object.");
}
*/

/*
void rb_simplesensel_setscandetail(t_simplesensel *x, t_symbol *s, long argc, t_atom *argv)
// check if device open first?
{
	if (x->current_device != 0xFF)
	{
		if (argc > 0) {
			post("Number of arguments: %d.\n", argc);
			if (argv[0].a_type == A_LONG) {
				post("Value of argument: %d.\n", atom_getlong(argv));
				switch (atom_getlong(argv)) {
				case 0:
					if (senselSetScanDetail(x->handle, SCAN_DETAIL_HIGH) == 0) {
						post("Scan detail successfully set.\n");
					}
					else error("Couldn't set scan detail.");
					break;
				case 1:
					if (senselSetScanDetail(x->handle, SCAN_DETAIL_MEDIUM) == 0) {
						post("Scan detail successfully set.\n");
					}
					else error("Couldn't set scan detail.");
					break;
				case 2:
					if (senselSetScanDetail(x->handle, SCAN_DETAIL_LOW) == 0) {
						post("Scan detail successfully set.\n");
					}
					else error("Couldn't set scan detail.");
					break;
				default:
					error("Wrong argument value (should be 0, 1 or 2.)\n");
					break;
				}
			}
			else post("Wrong argument type.\n");
		}
		else error("Please provide an argument to the \"setscandetail\" message.\n");
	}
	else error("No Sensel Morph associated with this object.");
}
*/

/*
void rb_simplesensel_sleeptime(t_simplesensel *x, long sleeptime)
{
	if (sleeptime<2)
		sleeptime = 2;
	x->a_sleeptime = sleeptime;			// no need to lock since we are readonly in worker thread
}
*/

